// Hash functions 

#include "types.h"
#include "xtea.h"
#include "idea.h"
#include "hash_func.h"
#include "large_math.h"
#include "random.h"
#include "diffie_helman.h"
#include "serialize.h"

static void hash_block(u32 *hash, u8 const* data, u32 data_length_bytes, void (*hash_function)(u32 *,u32 *const));
static void mix_128(u32 *x);
static void unmix_128(u32 *x);


//----------------------------
// Nonlinear keyless mixing
// Minimum 3 rounds over data to achieve diffusion
//


void mix_data(u32 *x, const u32 len, const u32 rounds)
{
u32 cnt = 0;

for(u32 i = 0; i < rounds; i++)
 {
 for(u32 j = 1; j < len; j++) 
  {
  x[j] += k_mix(x[j-1] + cnt);
  cnt++;
  }
 x[0] += k_mix(x[len-1] + cnt);
 cnt++; 
 } 
}


void unmix_data(u32 *x, const u32 len, const u32 rounds)
{
u32 cnt = len*rounds;    

for(u32 i = 0; i < rounds; i++)
 {
 cnt--;
 x[0] -= k_mix(x[len-1] + cnt);
   
 for(u32 j = len-1; j > 0; j--)
  {
  cnt--;  
  x[j] -= k_mix(x[j-1] + cnt);
  }
 }
}


void xor_data(u32 *x, u32 const *xor_data, const u32 len)
{
for(u32 i = 0; i < len; i++) x[i] ^= xor_data[i];
}

/*
static const u32 mix_128_rounds = 4;
//
static void mix_128(u32 *x)
{

for(u32 i = 0; i < mix_128_rounds; i++)
 {
 x[1] += k_mix(x[0]);
 x[2] += k_mix(x[1]);
 x[3] += k_mix(x[2]);
 x[0] += k_mix(x[3]);
 }
}
//-----------------------------------
// Inverse mix (for testing)
//
static void unmix_128(u32 *x)
{
for(u32 i = 0; i < mix_128_rounds; i++)
 {
 x[0] -= k_mix(x[3]);
 x[3] -= k_mix(x[2]);
 x[2] -= k_mix(x[1]);
 x[1] -= k_mix(x[0]);
 }
}


void X_64(u32 *hash, u32 const* data)
{
u32 key[4]; // 128-bit key

key[0] = data[0];
key[2] = data[1];
key[1] = hash[0];
key[3] = hash[1];

XTEA_Encode(hash, key);
}

void I_64(u32 *hash, u32 const* data)
{
u32 key[4]; // 128-bit key

key[0] = data[0];
key[2] = data[1];
key[1] = hash[0];
key[3] = hash[1];

IDEA_Encode(hash, key);
}

//----------------------------------
// 128 bit hash, 64 bit data
// 
void I_128(u32 *hash, u32 const* data)
{
u32 key[4];
u32 tmp_data[2];

key[0] = data[0];
key[2] = data[1];
key[1] = hash[2];
key[3] = hash[3];

tmp_data[0] = hash[0];
tmp_data[1] = hash[1];

IDEA_Encode(tmp_data, key);

key[0] = hash[0];
key[2] = hash[1];
key[1] = data[0];
key[3] = data[1];

hash[0] ^= tmp_data[0];
hash[1] ^= tmp_data[1];

tmp_data[0] = hash[2];
tmp_data[1] = hash[3];

IDEA_Encode(tmp_data, key);

hash[2] ^= tmp_data[0];
hash[3] ^= tmp_data[1];
}

//----------------------------------
// 128 bit hash, 64 bit data
// 
void X_128(u32 *hash, u32 const* data)
{
u32 key[4];
u32 tmp_data[2];

key[0] = data[0];
key[2] = data[1];
key[1] = hash[2];
key[3] = hash[3];

tmp_data[0] = hash[0];
tmp_data[1] = hash[1];

XTEA_Encode(tmp_data, key);

key[0] = hash[0];
key[2] = hash[1];
key[1] = data[0];
key[3] = data[1];

hash[0] ^= tmp_data[0];
hash[1] ^= tmp_data[1];

tmp_data[0] = hash[2];
tmp_data[1] = hash[3];

mix_128(key);
XTEA_Encode(tmp_data, key);

hash[2] ^= tmp_data[0];
hash[3] ^= tmp_data[1];
}
*/
//----------------------------
// 256 bit hash, 64 bit data
// Weak (?)
//
void XI_256(u32 *hash, u32 const *data)
{
u32 key[4];
u32 tmp_data[8], lock[8];

for(u32 i = 0; i < 8; i++) lock[i] = tmp_data[i] = hash[i];


void (*f0)(u32 *, u32 const *) = XTEA_Encode;
void (*f1)(u32 *, u32 const *) = IDEA_Encode;

// Cyclic shift of inputs
// 2 layers of CBC
// Use data at first
// Alternating XTEA and IDEA
// 00 
// 0,d,6,4,2

// Data goes into the first block

key[0] = data[0];
key[1] = data[1];
key[2] = hash[6];
key[3] = hash[7];

// First XOR is just in case we need to add more rounds

tmp_data[0] ^= tmp_data[6];
tmp_data[1] ^= tmp_data[7];
 
f0(tmp_data + 0, key);

// 10
// 2,0,d,6,4

key[0] = hash[0];
key[1] = hash[1];
key[2] = data[0];
key[3] = data[1];

tmp_data[2] ^= tmp_data[0];
tmp_data[3] ^= tmp_data[1];

f0(tmp_data + 2, key);

// 20
// 4,2,0,d,6 

key[0] = hash[2];
key[1] = hash[3];
key[2] = hash[0];
key[3] = hash[1];

tmp_data[4] ^= tmp_data[2];
tmp_data[5] ^= tmp_data[3];

f0(tmp_data + 4, key);

// 30
// 6,4,2,0,d

key[0] = hash[4];
key[1] = hash[5];
key[2] = hash[2];
key[3] = hash[3];

tmp_data[6] ^= tmp_data[4];
tmp_data[7] ^= tmp_data[5];

f0(tmp_data + 6, key);

//-----
// 01

key[0] = hash[4];
key[1] = hash[5];
key[2] = hash[2];
key[3] = hash[3];

tmp_data[0] ^= tmp_data[6];
tmp_data[1] ^= tmp_data[7];

f1(tmp_data + 0, key);

// 11
key[0] = hash[6];
key[1] = hash[7];
key[2] = hash[4];
key[3] = hash[5];

tmp_data[2] ^= tmp_data[0];
tmp_data[3] ^= tmp_data[1];

f1(tmp_data + 2, key);

// 21
key[0] = data[0];
key[1] = data[1];
key[2] = hash[6];
key[3] = hash[7];

tmp_data[4] ^= tmp_data[2];
tmp_data[5] ^= tmp_data[3];

f1(tmp_data + 4, key);

// 31
key[0] = hash[0];
key[1] = hash[1];
key[2] = data[0];
key[3] = data[1];

tmp_data[6] ^= tmp_data[4];
tmp_data[7] ^= tmp_data[5];

f1(tmp_data + 6, key);

// Lock

for(u32 i = 0; i < 8; i++) hash[i] = tmp_data[i] ^ lock[i];
}


///----------------------------
// 256 bit hash, 64 bit data
// 

void hash_block_256(u32 *hash, u32 const *data)
{
XI_256(hash, data);
}

//------------------------------------------------------------
// Generate hash of arbitrary data length
// Initial value of hash must be initialized 
//
static void hash_block(u32 *hash, u8 const* data, u32 data_length_bytes, void (*hash_function)(u32 *,u32 const*))
{
u32 tmp[2];

// Just for sanity

tmp[0] = 0;
tmp[1] = 0; 

for(u32 ci = 0; ci < data_length_bytes; ci++)
 {
 tmp[1] <<= 8;
 tmp[1] |= tmp[0] >> 24;
 tmp[0] <<= 8;
 tmp[0] |= data[ci];
 
 if((ci & 0x07) == 0x07) // 8 bytes loaded
  {
  hash_function(hash, tmp);
  }
 }

// Pad last block with zeroes

u32 ci = data_length_bytes & 0x07;
if(ci)
 {
 ci = 8 - ci;   
 while(ci)
  {
  ci--;
  tmp[1] <<= 8;
  tmp[1] |= tmp[0] >> 24;
  tmp[0] <<= 8;
  }
 hash_function(hash, tmp);
 }
}



//------------------------------------------------------------
// Generate 64-bit hash of arbitrary data length
// using XTEA
// Initial value of hash must be initialized 
//
void Hash_Block_X_64(u32 *hash, u8 const* data, u32 data_length_bytes)
{
u32 data_input[2], lock[2];
u32 key[4]; 
u32 tweak = 0;

#define TWEAK_PRNG tweak = (((tweak << 5) - tweak) ^ 0x6A551EE9)

// Chaining mode with random tweak value
   
while(data_length_bytes)
 {
 data_length_bytes = fetch_dwords(data_input, 2, data, data_length_bytes);
 data += 2*sizeof(u32);

 lock[0] = data_input[0];
 lock[1] = data_input[1];
 key[0] = hash[0];
 key[1] = hash[1]; 
 key[2] = TWEAK_PRNG; 
 key[3] = TWEAK_PRNG;

 XTEA_Encode(data_input, key);

 hash[0] = data_input[0] ^ lock[0];
 hash[1] = data_input[1] ^ lock[1];
 }

#undef TWEAK_PRNG
}


//--------------------

COLLAPSE::COLLAPSE(u32 hash_size_x32)
{
size_x32 = hash_size_x32;
iter_per_block = size_x32/2;

zero(tweak.seed, 2);


data_input = new u32[size_x32];
lock_bar = new u32[size_x32];
}

COLLAPSE::~COLLAPSE()
{
COLLAPSE::reset();
delete[] lock_bar;
delete[] data_input;
}

void COLLAPSE::reset()
{
cheap_random_number(data_input, size_x32);
cheap_random_number(lock_bar, size_x32);
cheap_random_number(tweak.seed, 2);
reset_stack();
}





void COLLAPSE::hash(u32 *hash, void const *data, u32 data_block_size_bytes)
{

COLLAPSE::process_hash(hash, data, data_block_size_bytes, // current length
                            data_block_size_bytes, // full length
                            NULL, // tweak
                            1); // initialize
   
}
                

void COLLAPSE::start_hash(u32 *hash, void const *data, u32 data_block_size_bytes,
                          const u32 total_size_bytes, COLLAPSE_TWEAK_T *tweak_ptr)
{

COLLAPSE::process_hash(hash, data, data_block_size_bytes, total_size_bytes, tweak_ptr, 1);


}

void COLLAPSE::continue_hash(u32 *hash, void const *data, u32 data_block_size_bytes,
                             COLLAPSE_TWEAK_T *tweak_ptr)
{

COLLAPSE::process_hash(hash, data, data_block_size_bytes, data_block_size_bytes, tweak_ptr, 0);

}


void COLLAPSE::process_hash(u32 *hash, 
                            void const *data, u32 current_data_block_size_bytes, 
                            u32 total_length_bytes,
                            COLLAPSE_TWEAK_T *tweak_ptr,
                            int initialize_hash)
   

{
u8 const* data_ptr = (u8 const*)data; 

// Initialize


if(initialize_hash)
 {   
 hash[0] = current_data_block_size_bytes;
 hash[1] = total_length_bytes;
 
 zero(tweak.seed, 2);

 for(u32 i = 2; i < size_x32; i++) hash[i] = lcg64(tweak.seed);

 }
else
 {
 if(tweak_ptr != NULL) tweak = *tweak_ptr; 
 // otherwise continue with current value of tweak
 }

// Main loop

while(current_data_block_size_bytes)
 {
 // Read data   
 
 current_data_block_size_bytes = fetch_dwords(data_input, size_x32, data_ptr, current_data_block_size_bytes);
 
 data_ptr += size_x32*sizeof(u32);

 copy(lock_bar, data_input, size_x32);
 
 // Encrypt: note that hash is destroyed in this function
  
 COLLAPSE::encrypt(data_input, hash, &tweak);
 
 // Lock up
 
 for(u32 i = 0; i < size_x32; i++) hash[i] = data_input[i] ^ lock_bar[i];
 } // data_size_bytes

// Remember the state

if(tweak_ptr != NULL) *tweak_ptr = tweak;

// Do not reset; retain the state
}

//--------------------------------------------------------------
// Block cipher
// Note: key is destroyed in process
//
void COLLAPSE::encrypt(u32 *data, u32 *key, COLLAPSE_TWEAK_T *tw)
{

 // Iterate through block
  
 for(u32 i = 0; i < iter_per_block; i++)
  {
  // MIX layer

  xor_data(data, key, size_x32/2);
  for(u32 j = size_x32/2; j < size_x32; j++) data[j] ^= lcg64(tw->seed);
  mix_data(data, size_x32, collapse_mix_rounds);

  xor_data(data, key + size_x32/2, size_x32/2);
  for(u32 j = size_x32/2; j < size_x32; j++) data[j] ^= lcg64(tw->seed);
  mix_data(data, size_x32, collapse_mix_rounds);

  // IDEA layer
  
  for(u32 j = 0; j < size_x32; j += 2)
   {
   u32 idea_key[4];

   idea_key[0] = key[j + 0];
   idea_key[1] = key[j + 1];
   idea_key[2] = lcg64(tw->seed);
   idea_key[3] = lcg64(tw->seed);

   IDEA_Encode(data + j, idea_key); 
   }
  
  // XXTEA layer

  // We switch XTEA key so it would match to IDEA key of different block every time

  u32 xxtea_key[4];

  xxtea_key[0] = key[2*i + 0];
  xxtea_key[1] = key[2*i + 1];
  xxtea_key[2] = lcg64(tw->seed);
  xxtea_key[3] = lcg64(tw->seed);

  XXTEA_Encode(data, size_x32, xxtea_key, collapse_xxtea_rounds);

  // Key mod

  for(u32 j = 0; j < size_x32; j++) key[j] ^= lcg64(tw->seed);

  mix_data(key, size_x32, collapse_mix_rounds);
  } // iter_per_block

}
//-----------------------------------------------------------
// Decrypt operation
// Just for testing
//
void COLLAPSE::decrypt(u32 *data, u32 *key, COLLAPSE_TWEAK_T *tw)
{
// MIX uses size_x32 tweaks
// IDEA uses size_x32 tweaks
// XXTEA uses 2 tweaks
// Keymod uses size_x32 tweaks


u32 tweak_size_x32  = iter_per_block*(size_x32 + size_x32 + 2 + size_x32);

u32 *tweak_value = new u32[tweak_size_x32];

for(u32 i = 0; i < tweak_size_x32; i++) tweak_value[i] = lcg64(tw->seed); 

// Iterate through block

u32 *tweak_ptr = tweak_value + tweak_size_x32 - 1;
  
u32 i = iter_per_block;
 
 while(i--)
  {
  // Key mod

  unmix_data(key, size_x32, collapse_mix_rounds);
  for(u32 j = size_x32; j > 0; j--) key[j-1] ^= *tweak_ptr--;


  // XXTEA layer

  u32 xxtea_key[4];

  xxtea_key[0] = key[2*i + 0];
  xxtea_key[1] = key[2*i + 1];

  xxtea_key[3] = *tweak_ptr--;
  xxtea_key[2] = *tweak_ptr--;
  
  XXTEA_Decode(data, size_x32, xxtea_key, collapse_xxtea_rounds); 
  
  // IDEA layer

  u32 j = size_x32;

  while(j)
   {
   j -= 2;  
   u32 idea_key[4];

   idea_key[0] = key[j + 0];
   idea_key[1] = key[j + 1];
   
   idea_key[3] = *tweak_ptr--;
   idea_key[2] = *tweak_ptr--;
   
   IDEA_Decode(data + j, idea_key); 
   }

 
  // MIX layer

  unmix_data(data, size_x32, collapse_mix_rounds);

  xor_data(data, key + size_x32/2, size_x32/2);
  for(u32 j = size_x32 - 1; j >= size_x32/2; j--) data[j] ^= *tweak_ptr--;

  unmix_data(data, size_x32, collapse_mix_rounds);
  xor_data(data, key, size_x32/2);
  for(u32 j = size_x32 - 1; j >= size_x32/2; j--) data[j] ^= *tweak_ptr--;
  
  } // iter_per_block

delete[] tweak_value;
}


//-----------------------

DH_HASH::DH_HASH()
{
collapse = new COLLAPSE(fingerprint_length_x32);
}

DH_HASH::~DH_HASH()
{
DH_HASH::reset();
reset_stack(); 
delete collapse;
}

void DH_HASH::reset()
{
collapse->reset();
reset_stack();   
}

void DH_HASH::hash(u32 *hash, u32 const* g, u32 const* y, u32 const* p)
{
COLLAPSE_TWEAK_T collapse_tweak_t;
    
collapse->start_hash(hash, g, dh_length_x32*sizeof(u32), 3*dh_length_x32*sizeof(u32), &collapse_tweak_t);

collapse->continue_hash(hash, y, dh_length_x32*sizeof(u32), &collapse_tweak_t);

collapse->continue_hash(hash, p, dh_length_x32*sizeof(u32), &collapse_tweak_t);

DH_HASH::reset();
}



//===================== <END> =====================
